/*
 * de405.c - readout DE405/406 ephemerides
 *  Copyright (C) 2002-2005 Noriaki TANAKA (dtana@startide.jp)
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
 *
 * $Id: de405.c,v 1.61 2005/05/08 11:02:49 dtana Exp $
 */
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <math.h>
#include "de405.h"

#ifndef EFTYPE
#define EFTYPE EINVAL
#endif

static int
freadr(void *p, size_t size, FILE *fp)
{
	int n;

	if ((n = fread(p, size, 1, fp)) == 1)
		return 0;
	if (n == 0)
		errno = EFTYPE;
	return -1;
}

/* read big endian 32bit value */
static int
freadl(long *d, FILE *fp)
{
	unsigned char b[4];

	if (freadr(b, 4, fp) < 0)
		return -1;
	*d = (long)(((unsigned long)b[0] << 24) |
	    ((unsigned long)b[1] << 16) |
	    ((unsigned long)b[2] << 8) | b[3]);
	return 0;
}

static int bigendian = 0;

/* read big endian double */
static int
freadd(double *d, FILE *fp)
{
	if (bigendian) {
		if (freadr(d, sizeof(double), fp) < 0)
			return -1;
	} else {
		union dbuf {
			double d;
			unsigned char be[8], le[8];
		} b, *r = (union dbuf *)d;

		if (freadr(&b, sizeof(double), fp) < 0)
			return -1;
		r->le[0] = b.be[7];
		r->le[1] = b.be[6];
		r->le[2] = b.be[5];
		r->le[3] = b.be[4];
		r->le[4] = b.be[3];
		r->le[5] = b.be[2];
		r->le[6] = b.be[1];
		r->le[7] = b.be[0];
	}
	return 0;
}

static int
fread_itp(struct ipt_t *p, FILE *fp)
{
	long rb;

	if (freadl(&rb, fp) < 0)
		return -1;
	/* changed offset to zero origin */
	p->offset = rb - 1;

	if (freadl(&rb, fp) < 0)
		return -1;
	p->ncf = rb;

	if (freadl(&rb, fp) < 0)
		return -1;
	p->na = (double)rb;
	return 0;
}

/* number of elements per record */
static int
nelements(int de)
{
	struct {
		int de, ndata;
	} data[] = {
		{200, 1652 / 2},
		{403, 2036 / 2},
		{404, 1456 / 2},
		{405, 2036 / 2},
		{406, 1456 / 2},
		{  0,	     0},
	}, *p;

	for (p = data; p->de != 0; p++)
		if (p->de == de)
			return p->ndata;

	/* fprintf(stderr, "%d: unknown DE number\n", de); */
	errno = EFTYPE;
	return -1;
}

void
de405_close(DE405 *p)
{
	int ekeep = errno;;

	if (p == NULL)
		return;
	if (p->fp != NULL)
		fclose(p->fp);
	free(p->eb);
	free(p);
	errno = ekeep;
}

static void
strchomp(char *b)
{
	while (*b && isspace(b[strlen(b) - 1]))
		b[strlen(b) - 1] = 0;
}

DE405 *
de405_open(const char *path)
{
	int i, k;
	DE405 *hp;
	long rb;

	/* check byteorder */
	bigendian = (k = 1, *(char *)&k != 1);

	if ((hp = malloc(sizeof(DE405))) == NULL)
		return NULL;
	hp->eb = NULL;
	if ((hp->fp = fopen(path, "rb")) == NULL)
		return NULL;

	/* title strings (three lines) */
	for (i = 0; i < 3; i++) {
		if (freadr(hp->title[i], 84, hp->fp) < 0)
			goto error_exit;
		hp->title[i][84] = 0;
		strchomp(hp->title[i]);
	}

	/* name of constants */
	for (i = 0; i < 400; i++) {
		if (freadr(hp->constant[i].name, 6, hp->fp) < 0)
			goto error_exit;
		hp->constant[i].name[6] = 0;
		strchomp(hp->constant[i].name);
	}

	/* time span for ephemerides */
	if (freadd(&hp->bjd, hp->fp) < 0	||
	    freadd(&hp->ejd, hp->fp) < 0	||
	    freadd(&hp->step, hp->fp) < 0)
		goto error_exit;

	/* other constants */
	if (freadl(&rb, hp->fp) < 0)
		goto error_exit;
	hp->nconstants = (int)rb;

	if (freadd(&hp->au, hp->fp) < 0 ||
	    freadd(&hp->emratio, hp->fp) < 0)
		goto error_exit;
	hp->AU = 1.0 / hp->au;

	/* interp pointers */
	for (k = 0; k < 12; k++)
		if (fread_itp(&hp->ipt[k], hp->fp) < 0)
			goto error_exit;

	/* DE number */
	if (freadl(&rb, hp->fp) < 0)
		goto error_exit;
	hp->denum = (int)rb;
	if ((hp->ndata = nelements(hp->denum)) < 0)
		goto error_exit;

	/* interp pointers for libration */
	if (fread_itp(&hp->ipt[12], hp->fp) < 0)
		goto error_exit;

	/* ipt consistency check */
	for (k = 0; k < 13; k++) {
		hp->ipt[k].ncm = k == DE_NUTATIONS ? 2: 3;

		if (hp->ipt[k].ncf == 0)
			continue;
		if (hp->ipt[k].offset < 0 || hp->ndata < hp->ipt[k].offset +
		    (int)hp->ipt[k].na * hp->ipt[k].ncm *hp->ipt[k].ncf) {
			errno = EFTYPE;
			return NULL;
		}
	}

	/* constants */
	if (fseek(hp->fp, hp->ndata * sizeof(double), SEEK_SET) < 0)
		goto error_exit;
	for (i = 0; i < 400; i++)
		if (freadd(&hp->constant[i].c, hp->fp) < 0)
			goto error_exit;

	if ((hp->eb = malloc(sizeof(double) * hp->ndata)) == NULL)
		goto error_exit;
	hp->nr = (long)-1;
	return hp;

 error_exit:
	de405_close(hp);
	return NULL;
}

static int
record(DE405 *hp, long nr)
{
	int k;
	long pt;

	pt = (nr + 2) * hp->ndata * sizeof(double);
	if (fseek(hp->fp, pt, SEEK_SET) < 0)
		return -1;

	for (k = 0; k < hp->ndata; k++)
		if (freadd(hp->eb + k, hp->fp) < 0)
			return -1;
	return 0;
}

#define NCF 18

/* interpolation */
static int
chebyshev(const double eb[], double ft, double step,
    int ncf, int ncm, double na, double p[3], double v[3])
{
	int i, j;
	double cp[NCF], nat, t;
	const double *tp;

	assert(0 < ncf && ncf < NCF);
	assert(0.0 <= ft && ft < 1.0);
	assert(ncm == 3 || ncm == 2); /* 2 for Nutations */

	nat = na * ft;

	/* base index for coefficients */
	eb += (int)nat * ncm * ncf;
	/* -1.0 <= t < 1.0 */
	t = 2.0 * (nat - floor(nat)) - 1.0;

	/* position */
	cp[0] = 1.0;
	cp[1] = t;
	for (i = 2; i < ncf; i++)
		cp[i] = 2.0 * cp[i - 1] * t - cp[i - 2];

	if (p != NULL) {
		for (i = 0, tp = eb; i < ncm; i++, tp += ncf)
			for (j = ncf - 1, p[i] = 0.0; 0 <= j; j--)
				p[i] += cp[j] * tp[j];
	}

	/* velocity */
	if (v != NULL) {
		double cv[NCF], rv;

		cv[0] = 0.0;
		cv[1] = 1.0;
		cv[2] = 4.0 * t;
		for (i = 3; i < ncf; i++)
			cv[i] = 2.0 * (cv[i - 1] * t + cp[i - 1]) - cv[i - 2];

		rv = 2.0 * na / step;
		for (i = 0, tp = eb; i < ncm; i++, tp += ncf) {
			for (j = ncf - 1, v[i] = 0.0; 1 <= j; j--)
				v[i] += cv[j] * tp[j];
			v[i] *= rv;
		}
	}
	return 0;
}

int
de405_ephem(DE405 *hp, double jd, int n, double p[3], double v[3])
{
	long nr;
	double t;

	if (jd < hp->bjd || hp->ejd <= jd) {
		errno = EINVAL;
		return -1;
	}

	t = (jd - hp->bjd) / hp->step;
	if ((nr = (long)t) != hp->nr) {
		if (record(hp, nr) < 0) {
			hp->nr = (long)-1;
			return -1;
		}
		hp->nr = nr;
	}

	if (hp->ipt[n].ncf == 0) {
		errno = EINVAL;
		return -1;
	}
	t -= floor(t);
	return chebyshev(hp->eb + hp->ipt[n].offset, t, hp->step,
	    hp->ipt[n].ncf, hp->ipt[n].ncm, hp->ipt[n].na, p, v);
}

/* helper */
int
de405_ephem_earth(DE405 *hp, double jd, double p[3], double v[3])
{
	int k;
	double em[6], mg[6], a;

	if (de405_ephem(hp, jd,
		DE_EARTH_MOON_BARYCENTER, em, em + 3) < 0 ||
	    de405_ephem(hp, jd, DE_MOON, mg, mg + 3) < 0)
		return -1;

	a = 1.0 / (hp->emratio + 1.0);

	if (p != NULL)
		for (k = 0; k < 3; k++)
			p[k] = em[k] - a * mg[k];
	if (v != NULL)
		for (k = 0; k < 3; k++)
			v[k] = em[k + 3] - a * mg[k + 3];
	return 0;
}

static int
stricmp_(const char *a, const char *b)
{
	while (*a && *b && (islower(*a) ? toupper(*a): *a) ==
	    (islower(*b) ? toupper(*b): *b))
		a++, b++;
	return *a - *b;
}

double
de405_constant(DE405 *hp, const char *name)
{
	int k;

	for (k = 0; k < hp->nconstants; k++)
		if (stricmp_(name, hp->constant[k].name) == 0)
			return hp->constant[k].c;
	return HUGE_VAL;
}

#ifdef COMMAND

#define RFORMAT "%17.13f"
#define SFORMAT "%19.12e"
#define TFORMAT "%25.18e"

static void
parray(const double r[], int n, const char *format)
{
	int k;

	for (k = 0; k < n; k++)
		printf(format, r[k]), printf(" ");
	printf("\n");
}

static void
pvector(const double r[3])
{
	printf(RFORMAT ": ",  sqrt(r[0]*r[0] + r[1]*r[1] + r[2]*r[2]));
	parray(r, 3, RFORMAT);
}

static int
dumpdata(DE405 *hp, double jd)
{
	int k;
	double r[6];

	for (k = 0; k < 13; k++) {
		if (de405_ephem(hp, jd, k, r, r + 3) < 0)
			continue;

		printf("%2d: ", k);
		if (k == DE_NUTATIONS) {
			parray(r, 2, SFORMAT);
		} else if (k == DE_LIBRATIONS) {
			parray(r, 3, SFORMAT);
		} else {
			int i;

			for (i = 0; i < 6; i++)
				r[i] *= hp->AU;
			pvector(r);
			printf("    ");
			pvector(r + 3);
		}
		printf("\n");
	}
	return 0;
}

static int
showdata(DE405 *hp, double jd, int n)
{
	double r[6];

	if (n < 0 || 12 < n)
		fprintf(stderr, "%d: number of object out of range"
		    "0 <= n < 12\n", n), exit(1);

	r[2] = r[5] = 0;
	if (de405_ephem(hp, jd, n, r, r + 3) < 0)
		return -1;

	parray(r,     3, TFORMAT);
	parray(r + 3, 3, TFORMAT);
	return 0;
}

void
print_header(DE405 *hp, FILE *fp)
{
	int i;

	fprintf(fp, "DE%d\n", hp->denum);
	fprintf(fp, "\n");

	fprintf(fp, "data/record: %5d (%04X)\n", hp->ndata, hp->ndata);

	for (i = 0; i < hp->nconstants; i++)
		fprintf(fp, "%-6.6s: %24.15g\n",
		    hp->constant[i].name, hp->constant[i].c);


	fprintf(fp, "begin : %24.15g\n", hp->bjd);
	fprintf(fp, "end   : %24.15g\n", hp->ejd);
	fprintf(fp, "step  : %24.15g\n", hp->step);

	fprintf(fp, "AU/km : %24.15g\n", hp->au);
	fprintf(fp, "Me/Mm : %24.15g\n", hp->emratio);

	for (i = 0; i < 13; i++)
		fprintf(fp, "ipt[%2d] %3d %3d %3d\n", i,
		    (int)hp->ipt[i].offset,
		    (int)hp->ipt[i].ncf,
		    (int)hp->ipt[i].na);
}

static void
usage(void)
{
	fprintf(stderr, "Usage: de405 [-f <path>] <JD> [n ...]\n");
	exit(1);
}

#ifdef IEEEFP
#include <ieeefp.h>
static void
ieeefp()
{
	fpsetmask(0x7f);
	fpsetprec(FP_PE);
}
#else
#define ieeefp()
#endif

extern int getopt(), optind;
extern char *optarg;

int
main(int argc, char **argv)
{
	const char *path;
	int optc;
	double jd;
	DE405 *hp;

	path = getenv("DE405");

	while ((optc = getopt(argc, argv, "f:")) != EOF)
		switch (optc) {
		case 'f':
			path = optarg;
			break;
		default:
			usage();
		}
	ieeefp();

	if (path == NULL) {
		fputs("data file not defined.\n", stderr);
		fputs("define -f <path> or \"DE405\" "
		    "environment variable\n", stderr);
		exit(1);
	}

	if ((hp = de405_open(path)) == NULL)
		perror(path), exit(1);

	if (optind == argc) {
		print_header(hp, stdout);
		return 0;
	}

	jd = atof(argv[optind++]);
	if (jd < hp->bjd || hp->ejd <= jd)
		fprintf(stderr, "%f: out of date\n", jd), exit(1);

	if (optind == argc) {
		printf("Data: DE%d (%s)\n",
		    hp->denum, hp->denum < 200 ? "B1950.0": "J2000.0");
		printf("Time-span: %.1f <= JD < %.1f\n", hp->bjd, hp->ejd);
		printf("JD: %f\n", jd);
		printf("\n");
		printf("Barycentric Position (in AU), Velocity (AU/day):\n");
		if (dumpdata(hp, jd) < 0)
			perror(path), exit(1);
	} else {
		while (optind != argc)
			if (showdata(hp, jd, atoi(argv[optind++])) < 0)
				perror(path), exit(1);
	}
	return 0;
}

#endif
