/*
 * testeph.c - testeph.f compatible DE405/406 regression test
 *  Copyright (C) 2002, 2004 Noriaki TANAKA (dtana@startide.jp)
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: testeph.c,v 1.14 2004/02/14 12:56:26 dtana Exp $
 */
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include "de405.h"

enum {
	MERCURY = 1,
	VENUS,
	EARTH,
	MARS,
	JUPITER,
	SATURN,
	URANUS,
	NEPTUNE,
	PLUTO,
	MOON,
	SUN,
	SOLAR_SYSTEM_BARYCENTER,
	EARTH_MOON_BARYCENTER,
	NUTATIONS,
	LIBRATIONS
};

int
barycentric(DE405 *p, double jd, int code, double r[6])
{
	int n, k;

	if (code == EARTH)
		return de405_ephem_earth(p, jd, r, r + 3);

	if (code == MOON) {
		double mg[6], ea[6];

		if (de405_ephem(p, jd, DE_MOON, mg, mg + 3) < 0 ||
		    de405_ephem_earth(p, jd, ea, ea + 3) < 0)
			return -1;
		for (k = 0; k < 6; k++)
			r[k] = ea[k] + mg[k];
		return 0;
	}

	if (code == SOLAR_SYSTEM_BARYCENTER) {
		for (k = 0; k < 6; k++)
			r[k] = 0.0;
		return 0;
	}

	n = code - 1;

	if (code == EARTH_MOON_BARYCENTER)
		n = DE_EARTH_MOON_BARYCENTER;
	if (code == SUN)
		n = DE_SUN;
	return de405_ephem(p, jd, n, r, r + 3);
}

int
ephem(DE405 *p, double jd, int target, int center, double r[6])
{
	int k;
	double tr[6], cr[6];

	if (target == NUTATIONS)
		return de405_ephem(p, jd, DE_NUTATIONS, r, r + 2);
	if (target == LIBRATIONS)
		return de405_ephem(p, jd, DE_LIBRATIONS, r, r + 3);

	if (barycentric(p, jd, target, tr) < 0)
		return -1;
	if (barycentric(p, jd, center, cr) < 0)
		return -1;
	for (k = 0; k < 6; k++)
		r[k] = (tr[k] - cr[k]) * p->AU;
	return 0;
}

#define DELTA 1E-13

int
testeph(FILE *fp, const char *data, const char *testpo)
{
	FILE *fp1;
	char b[BUFSIZ];
	long line;
	DE405 *hp;

	if ((hp = de405_open(data)) == NULL)
		perror(data), exit(1);

	if ((fp1 = fopen(testpo, "r")) == NULL)
		perror(testpo), exit(1);
	while (fgets(b, BUFSIZ, fp1) && strncmp(b, "EOT", 3))
		;
	if (ferror(fp1) || feof(fp1))
		fprintf(stderr, "%s: no test record\n", testpo), exit(1);

	fprintf(fp, "  line          jed  t  c  x        difference\n");
	for (line = 0; fgets(b, BUFSIZ, fp1) != NULL;) {
		double td, xi, delta, r[6];
		int target, center, ne;

		if (sscanf(b + 15, "%lg %d %d %d %lg",
			&td, &target, &center, &ne, &xi) != 5) {
			fprintf(stderr, "%s: invalid record\n", testpo),
			    exit(1);
		}
		if (td < hp->bjd || hp->ejd <= td)
			continue;

		if (ephem(hp, td, target, center, r) < 0)
			perror(data), exit(1);

		delta = r[ne - 1] - xi;
		fprintf(fp, "%+6ld %+12.2f %2d %2d %2d %+30.18g\n",
		    ++line, td, target, center, ne, delta);

		if (DELTA <= fabs(delta)) {
			fprintf(stderr,
			    "ERROR: %6ld %20.13g %20.13g\n", line, td, delta);
			exit(1);
		}
	}
	if (ferror(fp1))
		perror(testpo), exit(1);

	fclose(fp1);
	fprintf(stderr, "%ld records processed\n", line);

	return 0;
}

int
main(int argc, char **argv)
{
	if (argc != 3) {
		fprintf(stderr, "Usage: %s <data> <testdata>\n", argv[0]);
		exit(1);
	}
	return testeph(stdout, argv[1], argv[2]);
}
